from typing import Any, Dict
from .sensor import Sensor
from .const import Identifier
from .const import SMATagList


ennexosSensorProfiles = {
    "Sunny Tripower X ": [
        Sensor("Coolsys.Inverter.TmpVal.1", Identifier.temp_a, factor=1, unit="°C"),
        Sensor("Coolsys.Inverter.TmpVal.2", Identifier.temp_b, factor=1, unit="°C"),
        Sensor("Coolsys.Inverter.TmpVal.3", Identifier.temp_c, factor=1, unit="°C"),
        Sensor("DcMs.Amp.1", Identifier.pv_current_a, factor=1, unit="A"),
        Sensor("DcMs.Amp.2", Identifier.pv_current_b, factor=1, unit="A"),
        Sensor("DcMs.Amp.3", Identifier.pv_current_c, factor=1, unit="A"),
        Sensor("DcMs.TotDcEnCntWh.1", None, factor=1, unit=None), # Energy released by string [A]
        Sensor("DcMs.TotDcEnCntWh.2", None, factor=1, unit=None),
        Sensor("DcMs.TotDcEnCntWh.3", None, factor=1, unit=None),
        Sensor("DcMs.Vol.1", Identifier.pv_voltage_a, factor=1, unit="V"),
        Sensor("DcMs.Vol.2", Identifier.pv_voltage_b, factor=1, unit="V"),
        Sensor("DcMs.Vol.3", Identifier.pv_voltage_c, factor=1, unit="V"),
        Sensor("DcMs.Watt.1", Identifier.pv_power_a, factor=1, unit="W"),
        Sensor("DcMs.Watt.2", Identifier.pv_power_b, factor=1, unit="W"),
        Sensor("DcMs.Watt.3", Identifier.pv_power_c, factor=1, unit="W"),
        Sensor("GridGuard.Cntry", None, factor=1, unit=None), # Country standard set
        Sensor("GridMs.A.phsA", Identifier.current_l1, factor=1, unit="A"),
        Sensor("GridMs.A.phsB", Identifier.current_l2, factor=1, unit="A"),
        Sensor("GridMs.A.phsC", Identifier.current_l3, factor=1, unit="A"),
        Sensor("GridMs.GriTyp", None, factor=1, unit=None), # Measurement.GridMs.GriTyp
        Sensor("GridMs.Hz", Identifier.frequency, factor=1, unit="Hz"),
        Sensor("GridMs.PhV.phsA", Identifier.voltage_l1, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsA2B", None, factor=1, unit=None), # Grid voltage phase L1 against L2
        Sensor("GridMs.PhV.phsB", Identifier.voltage_l2, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsB2C", None, factor=1, unit=None), # Grid voltage phase L2 against L3
        Sensor("GridMs.PhV.phsC", Identifier.voltage_l3, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsC2A", None, factor=1, unit=None), # Grid voltage phase L3 against L1
        Sensor("GridMs.TotA", Identifier.current_total, factor=1, unit="A"),
        Sensor("GridMs.TotPFEEI", None, factor=1, unit=None),  # EEI displacement power factor
        Sensor("GridMs.TotPFExt", None, factor=1, unit=None), # Excitation type of cos φ
        Sensor("GridMs.TotPFPrc", None, factor=1, unit=None), # Displacement power factor
        Sensor("GridMs.TotVA", Identifier.grid_apparent_power, factor=1, unit="VA"),
        Sensor("GridMs.TotVAr", Identifier.grid_reactive_power, factor=1, unit="var"),
        Sensor("GridMs.TotW", Identifier.grid_power, factor=1, unit="W"),
        Sensor("GridMs.TotW.Pv", Identifier.pv_power, factor=1, unit="W"),
        Sensor("GridMs.VA.phsA", Identifier.grid_apparent_power_l1, factor=1, unit="VA"),
        Sensor("GridMs.VA.phsB", Identifier.grid_apparent_power_l2, factor=1, unit="VA"),
        Sensor("GridMs.VA.phsC", Identifier.grid_apparent_power_l3, factor=1, unit="VA"),
        Sensor("GridMs.VAr.phsA", Identifier.grid_reactive_power_l1, factor=1, unit="var"),
        Sensor("GridMs.VAr.phsB", Identifier.grid_reactive_power_l2, factor=1, unit="var"),
        Sensor("GridMs.VAr.phsC", Identifier.grid_reactive_power_l3, factor=1, unit="var"),
        Sensor("GridMs.W.phsA", Identifier.power_l1, factor=1, unit="W"),
        Sensor("GridMs.W.phsB", Identifier.power_l2, factor=1, unit="W"),
        Sensor("GridMs.W.phsC", Identifier.power_l3, factor=1, unit="W"),
        Sensor("InOut.GI1", None, factor=1, unit=None), # Digital group input
        Sensor("Inverter.VArModCfg.PFCtlVolCfg.Stt", None, factor=1, unit=None),
        Sensor("Isolation.FltA", Identifier.insulation_residual_current, factor=1000, unit="mA"),
        Sensor("Isolation.LeakRis", None, factor=1, unit="kOhm"), # TODO "pv_isolation_resistance"
        Sensor("Metering.TotFeedTms", None, factor=1, unit=None), 
        Sensor("Metering.TotOpTms", None, factor=1, unit=None),
        Sensor("Metering.TotWhOut", Identifier.total_yield, factor=1000, unit="kWh"),
        Sensor("Metering.TotWhOut.Pv", Identifier.pv_gen_meter, factor=1000, unit="kWh"),
        Sensor("Operation.BckStt", None, factor=1, unit=None),
        Sensor("Operation.DrtStt", None, factor=1, unit=None),
        Sensor("Operation.Evt.Dsc", None, factor=1, unit=None),
        Sensor("Operation.Evt.EvtNo", None, factor=1, unit=None),
        Sensor("Operation.Evt.Msg", None, factor=1, unit=None),
        Sensor("Operation.EvtCntIstl", None, factor=1, unit=None),
        Sensor("Operation.EvtCntUsr", None, factor=1, unit=None),
        Sensor("Operation.GriSwCnt", None, factor=1, unit=None),
        Sensor("Operation.GriSwStt", None, factor=1, unit=None),
        Sensor("Operation.Health", Identifier.status, factor=1, unit=None, mapper=SMATagList),
        Sensor("Operation.HealthStt.Alm", None, factor=1, unit=None),
        Sensor("Operation.HealthStt.Ok", None, factor=1, unit=None),
        Sensor("Operation.HealthStt.Wrn", None, factor=1, unit=None),
        Sensor("Operation.OpStt", None, factor=1, unit=None),
        Sensor("Operation.PvGriConn", None, factor=1, unit=None),
        Sensor("Operation.RstrLokStt", None, factor=1, unit=None),
        Sensor("Operation.RunStt", None, factor=1, unit=None),
        Sensor("Operation.StandbyStt", None, factor=1, unit=None),
        Sensor("Operation.VArCtl.VArModAct", None, factor=1, unit=None),
        Sensor("Operation.VArCtl.VArModStt", None, factor=1, unit=None),
        Sensor("Operation.WMaxLimSrc", None, factor=1, unit=None),
        Sensor("Operation.WMinLimSrc", None, factor=1, unit=None),
        Sensor("PvGen.PvW", None, factor=1, unit=None), #  PV generation power
        Sensor("PvGen.PvWh", None, factor=1, unit=None), # Meter count and PV gen. meter
        Sensor("Spdwr.ComSocA.Stt", None, factor=1, unit=None), # Speedwire connection status of SMACOM A
        Sensor("SunSpecSig.SunSpecTx.1", None, factor=1, unit=None), #SunSpec life sign [1]
        Sensor("Upd.Stt", None, factor=1, unit=None),
        Sensor("WebConn.Stt", None, factor=1, unit=None), #  Status of the Webconnect functionality
        Sensor("Wl.AcqStt", None, factor=1, unit=None), # Status of WiFi scan
        Sensor("Wl.AntMod", None, factor=1, unit=None), #  WiFi antenna type
        Sensor("Wl.ConnStt", None, factor=1, unit=None), # WiFi connection status
        Sensor("Wl.SigPwr", None, factor=1, unit=None), # Signal strength of the selected network
        Sensor("Wl.SoftAcsConnStt", None, factor=1, unit=None), #Soft Access Point status
        Sensor("Setpoint.PlantControl.InOut.GO1", None, factor=1, unit=None)
    ],
    "SMA EV Charger ": [
        Sensor("ChaSess.WhIn", None, factor=1, unit=None), # charging_session_energy
        Sensor("Chrg.ModSw", None, factor=1, unit=None, mapper=SMATagList), # position_of_rotary_switch 4950 or 4718
        Sensor("GridMs.A.phsA", Identifier.current_l1, factor=1, unit="A"), #Netzstrom Phase L1
        Sensor("GridMs.A.phsB", Identifier.current_l2, factor=1, unit="A"),
        Sensor("GridMs.A.phsC", Identifier.current_l3, factor=1, unit="A"),
        Sensor("GridMs.Hz", Identifier.frequency, factor=1, unit="Hz"),
        Sensor("GridMs.PhV.phsA", Identifier.voltage_l1, factor=1, unit="V"),#Netzspannung Phase L1
        Sensor("GridMs.PhV.phsB", Identifier.voltage_l2, factor=1, unit="V"),
        Sensor("GridMs.PhV.phsC", Identifier.voltage_l3, factor=1, unit="V"),
        Sensor("GridMs.TotPF", None, factor=1, unit=None),
        Sensor("GridMs.TotVA", Identifier.grid_apparent_power, factor=1, unit="VA"),
        Sensor("GridMs.TotVAr", Identifier.grid_reactive_power, factor=1, unit="var"),
        Sensor("InOut.GI1", None, factor=1, unit=None),
        Sensor("Metering.GridMs.TotWIn", Identifier.metering_power_absorbed, factor=1, unit="W"), #
        Sensor("Metering.GridMs.TotWIn.ChaSta", None, factor=1, unit=None), # same as Metering.GridMs.TotWIn
        Sensor("Metering.GridMs.TotWhIn", Identifier.metering_total_absorbed, factor=1, unit="Wh"), # charging_station_meter_reading
        Sensor("Metering.GridMs.TotWhIn.ChaSta", None, factor=1, unit=None), # same as Metering.GridMs.TotWhIn 
        Sensor("Operation.EVeh.ChaStt", Identifier.operating_status, factor=1, unit=None, mapper=SMATagList), # charging_session_status
        Sensor("Operation.EVeh.Health", Identifier.status, factor=1, unit=None, mapper=SMATagList), 
        Sensor("Operation.Evt.Msg", None, factor=1, unit=None),
        Sensor("Operation.Health", None, factor=1, unit=None), # same as EVeh.Health?
        Sensor("Operation.WMaxLimNom", None, factor=1, unit=None),
        Sensor("Operation.WMaxLimSrc", None, factor=1, unit=None),
        Sensor("Wl.AcqStt", None, factor=1, unit=None),
        Sensor("Wl.ConnStt", None, factor=1, unit=None),
        Sensor("Wl.SigPwr", None, factor=1, unit=None),
        Sensor("Wl.SoftAcsConnStt", None, factor=1, unit=None),
    ]
}



# based on https://github.com/Wired-Square/sma-query/blob/main/src/sma_query_sw/commands.py
from .const import Identifier
from .sensor import Sensor
from .const import SMATagList

commands = {
    "login": {"command": 0xFFFD040C, 
              "response": 0xFFFD040D,
              "first": 0x0030CB00,
              "last": 0x0030CB00,

                "registers": [
                    {"name": "susyid", "format": "H", "offset": 20},
                    {"name": "serial", "format": "uint","offset": 22},
                    {"name": "pkt_ID", "format": "H","offset": 40},
                    {"name": "cmdid", "format": "H","offset": 42},
                    {"name": "error", "format": "uint","offset": 36},
                ]   
              
        #               my $r_susyid = unpack("v*", substr $data, 20, 2);
        # my $r_serial = unpack("V*", substr $data, 22, 4);
        # my $r_pkt_ID = unpack("v*", substr $data, 40, 2);
        # my $r_cmd_ID = unpack("V*", substr $data, 42, 4);
        # my $r_error  = unpack("V*", substr $data, 36, 4);

              },
    "logoff": {"command": 0xFFFD010E, "response": 0xFFFD010F},
    "TypeLabel": {
        "command": 0x58000200,
        "response": 0x00821E00,
        "first": 0x00821E00,
        "last": 0x008220FF,
        "registers": [
            {"name": "inverter_class", "offset": 102, "mask": 0x00FFFFFF},
            {"name": "inverter_type", "format": "list", "offset": 142},
            #$temp = unpack("V*", substr $data, $i, 4);

        ],
    },
    "EnergyProduction": {
        "command": 0x54000200,
        "response": 0x00260100,
        "first": 0x00260100,
        "last": 0x002622FF,
    },
    "PVEnergyProduction": {
        "command": 0x54000200,
        "response": 0x0046C300,
        "first": 0x0046C300,
        "last": 0x0046C3FF,
    },
    "SpotDCPower": {
        "command": 0x53800200,
        "response": 0x00251E00,
        "first": 0x00251E00,
        "last": 0x00251EFF,
        "registers": [
            {"name": "spot_dc_power1", "format": "int", "offset": 62,
                     "sensor": Sensor("spot_dc_power1", Identifier.pv_power_a, factor=1, unit="W"),

             },
            {"name": "spot_dc_power2", "format": "int","offset": 90, 
                     "sensor": Sensor("spot_dc_power2", Identifier.pv_power_b, factor=1, unit="W"),
},
        ],
    },
    "SpotDCPower_3": {
        "command": 0x51000200,
        "response": 0x0046C200,
        "first": 0x0046C200,
        "last": 0x0046C2FF,
    },
    "SpotACPower": {
        "command": 0x51000200,
        "response": 0x00464000,
        "first": 0x00464000,
        "last": 0x004642FF,
    },
    "SpotACTotalPower": {
        "command": 0x51000200,
        "response": 0x00263F00,
        "first": 0x00263F00,
        "last": 0x00263FFF,
        "registers": [
            {
                "name": "SpotACTotalPower",
                "offset": 62,
                "format": "int",
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "SpotACTotalPower", Identifier.grid_power, factor=1, unit="W"
                ),
            }
        ],
    },
    "ChargeStatus": {
        "command": 0x51000200,
        "response": 0x00295A00,
        "first": 0x00295A00,
        "last": 0x00295AFF,
        "registers": [
            {
                "name": "ChargeStatus",
                "offset": 62,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "ChargeStatus", Identifier.battery_soc_total, unit="%"
                ),
            }
        ],
    },
    "SpotDCVoltage": {
        "command": 0x53800200,
        "response": 0x00451F00, #TODO 0x451F
        "first": 0x00451F00,
        "last": 0x004521FF,
        "registers": [
            {"name": "spot_dc_voltage1", "offset": 62, "sensor": Sensor("spot_dc_voltage1", Identifier.pv_voltage_a, factor=100, unit="V")},
            {"name": "spot_dc_voltage2", "offset": 90, "sensor": Sensor("spot_dc_voltage2", Identifier.pv_voltage_b, factor=100, unit="V")},
            {"name": "spot_dc_current1", "offset": 118, "sensor": Sensor("spot_dc_current1", Identifier.pv_current_a, factor=1000, unit="A")},
            {"name": "spot_dc_current2", "offset": 146, "sensor": Sensor("spot_dc_current2", Identifier.pv_current_b, factor=1000, unit="A")},
        ],
    },
    "SpotACVoltage": {
        "command": 0x51000200,
        "response": 0x00464800,
        "first": 0x00464800,
        "last": 0x004656FF,
    },
    "SpotACCurrent": {
        "command": 0x51000200,
        "response": 0x00465300,
        "first": 0x00465300,
        "last": 0x004655FF,
        "registers": [
            {
                "name": "SpotACCurrent_L1",
                "offset": 62,
                "format": "int",
                "sensor": Sensor(
                    "SpotACCurrent_L1", Identifier.current_l1, factor=1000, unit="A"
                ),
            },
            {
                "name": "SpotACCurrent_L2",
                "offset": 90,
                "format": "int",
                "sensor": Sensor(
                    "SpotACCurrent_L2", Identifier.current_l2, factor=1000, unit="A"
                ),
            },
            {
                "name": "SpotACCurrent_L3",
                "offset": 118,
                "format": "int",
                "sensor": Sensor(
                    "SpotACCurrent_L3", Identifier.current_l3, factor=1000, unit="A"
                ),
            },
        ],
    },
    "SpotACCurrentBackup": {
        "command": 0x51000200,
        "response": 0x40574600,
        "first": 0x40574600,
        "last": 0x405748FF,
    },
    # "BatteryInfo_TEMP": { ## TODO
    #     "command": 0x51000200,
    #     "response": 0x00495B00,
    #     "first": 0x00495B00,
    #     "last": 0x00495B10,
    # },
    "BatteryInfo_UDC": {
        "command": 0x51000200,
        "response": 0x00495C00,
        "first": 0x00495C00,
        "last": 0x00495C10,
        "registers": [
            {
                "name": "battery_voltage_a",
                "offset": 62,
                "sensor": Sensor(
                    "battery_voltage_a",
                    Identifier.battery_voltage_a,
                    unit="V",
                    factor=100,
                ),
            },
            {
                "name": "battery_voltage_b",
                "offset": 90,
                "sensor": Sensor(
                    "battery_voltage_b",
                    Identifier.battery_voltage_b,
                    unit="V",
                    factor=100,
                ),
            },
            {
                "name": "battery_voltage_c",
                "offset": 118,
                "sensor": Sensor(
                    "battery_voltage_c",
                    Identifier.battery_voltage_c,
                    unit="V",
                    factor=100,
                ),
            },
        ],
    },
    "BatteryInfo_IDC": {
        "command": 0x51000200,
        "response": 0x00495D00,
        "first": 0x00495D00,
        "last": 0x00495D10,
        "registers": [
            {
                "name": "battery_current_a",
                "format": "int",
                "offset": 62,
                "sensor": Sensor(
                    "battery_current_a",
                    Identifier.battery_current_a,
                    unit="A",
                    factor=1000,
                ),
            },
            {
                "name": "battery_current_b",
                "format": "int",
                "offset": 90,
                "sensor": Sensor(
                    "battery_current_b",
                    Identifier.battery_current_b,
                    unit="A",
                    factor=1000,
                ),
            },
            {
                "name": "battery_current_c",
                "offset": 118,
                "format": "int",
                "sensor": Sensor(
                    "battery_current_c",
                    Identifier.battery_current_c,
                    unit="A",
                    factor=1000,
                ),
            },
        ],
    },
    "BatteryInfo_Charge": {
        "command": 0x51000200,
        "response": 0x00496900,
        "first": 0x00496900,
        "last": 0x00496AFF,
        "registers": [
            {
                "name": "bat_p_charge",
                "offset": 62,
                "sensor": Sensor("bat_p_charge", Identifier.battery_power_charge_total, unit="W", factor=1)
            },
            {
                "name": "bat_p_discharge",
                "offset": 90,
                "sensor": Sensor("bat_p_discharge", Identifier.battery_power_discharge_total, unit="W", factor=1)
            },
        ],
    },
    "BatteryInfo_Capacity": {
        "command": 0x51000200,
        "response": 0x00696E00,
        "first": 0x00696E00,
        "last": 0x00696E10,
        "registers": [
            {
                "name": "BatteryInfo_Capacity",
                "offset": 62,
                "sensor": Sensor(
                    "BatteryInfo_Capacity", Identifier.battery_capacity_total, unit="%"
                ),
            }
        ],
    },
    "BatteryInfo": {
        "command": 0x51000200,
        "response": 0x00491E00,
        "first": 0x00491E00,
        "last": 0x00495DFF,
        "registers": [
            {
                "name": "TEMP BAT_CYCLES",
                "offset": 62,
                "sensor": Sensor(
                    "TEMP BAT_CYCLES", None, unit=None
                ),
            }
        ],
    },
    "BatteryInfo_3": {
        "command": 0x58020200,
        "response": 0x08822C00,
        "first": 0x08822C00,
        "last": 0x08822CFF,
    },
    "BatteryInfo_4": {
        "command": 0x58000200,
        "response": 0x00893700,
        "first": 0x00893700,
        "last": 0x008937FF,
        "registers": [
            {
                "name": "rated capacity",
                "offset": 78,
            },
        ],
    },
    "BatteryInfo_5": {
        "command": 0x58020200,
        "response": 0x00B18900,
        "first": 0x00B18900,
        "last": 0x00B189FF,
    },
    "SpotGridFrequency": {
        "command": 0x51000200,
        "response": 0x00465700,
        "first": 0x00465700,
        "last": 0x004657FF,
        "registers": [
            {
                "name": "grid_frequency",
                "offset": 62,
                "sensor": Sensor(
                    "grid_frequency", Identifier.frequency, factor=100, unit="Hz"
                ),
            }
        ],
    },
    "OperationTime": {
        "command": 0x54000200,
        "response": 0x00462E00,
        "first": 0x00462E00,
        "last": 0x00462FFF,
        "registers": [
            {
                "name": "OperationTime + OperationTime2",
                "offset": 62,
            }
        ],

    },
    "InverterTemperature": {
        "command": 0x52000200,
        "response": 0x00237700,
        "first": 0x00237700,
        "last": 0x002377FF,
        "format": "int",
        "registers": [
            {"name": "InverterTemperature", "offset": 62, "invalid": 0x80000000}
        ],
    },
    "MaxACPower": {
        "command": 0x51000200,
        "response": 0x00411E00,
        "first": 0x00411E00,
        "last": 0x004120FF,
        "registers": [
            {
                "name": "MaxACPower1",
                "offset": 62,
                "invalid": 0x80000000,
                #     "sensor": Sensor("battery_current_a", Identifier.battery_current_a, unit="A", factor=1000)
            },
            {
                "name": "MaxACPower2",
                "offset": 90,
                "invalid": 0x80000000,
                # "sensor": Sensor("battery_current_b", Identifier.battery_current_b, unit="A", factor=1000)
            },
            {
                "name": "MaxACPower3",
                "offset": 118,
                "invalid": 0x80000000,
                # "sensor": Sensor("battery_current_b", Identifier.battery_current_b, unit="A", factor=1000)
            },
        ],
    },
    "MaxACPower2": {
        "command": 0x51000200,
        "response": 0x00832A00,
        "first": 0x00832A00,
        "last": 0x00832AFF,
    },
    "GridRelayStatus": {
        "command": 0x51800200,
        "response": 0x00416400,
        "first": 0x00416400,
        "last": 0x004164FF,
        "registers": [
            {"name": "grid_relay_status", "offset": 62, "invalid": 0x80000000}
        ],
    },
    "BackupRelayStatus": {
        "command": 0x51800200,
        "response": 0x08412500,
        "first": 0x08412500,
        "last": 0x084125FF,
    },
    "GridConection": {
        "command": 0x51800200,
        "response": 0x0846A600,
        "first": 0x0846A600,
        "last": 0x0846A6FF,
        # "registers": [
        #     {"name": "GridConection", "format": "list", "offset": 62},
        # ],
    },
    "OperatingStatus": {
        "command": 0x51800200,
        "response": 0x08412B00,
        "first": 0x08412B00,
        "last": 0x08412BFF,
    },
    "GeneralOperatingStatus": {
        "command": 0x51800200,
        "response": 0x08412800,
        "first": 0x08412800,
        "last": 0x084128FF,
        #"registers": [
        #    {"name": "GeneralOperatingStatus", "format": "list", "offset": 142},
        #],
    },
    "WaitingTimeUntilFeedIn": {
        "command": 0x51000200,
        "response": 0x00416600,
        "first": 0x00416600,
        "last": 0x004166FF,
        "registers": [
            {"name": "WaitingTimeUntilFeedIn", "offset": 62, "invalid": 0x80000000}
        ],
    },
    "DeviceStatus": {
        "command": 0x51800200,
        "response": 0x00214800,
        "first": 0x00214800,
        "last": 0x002148FF,
        "registers": [
            {   
                # TODO Schleife
                "name": "inverter_status",
                "offset": 62,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "inverter_status", Identifier.status, factor=1, unit=None, mapper = SMATagList
                ),
            }
        ],
    },
    "SpotBatteryLoad": {
        "command": 0x54000200,
        "response": 0x00496700,
        "first": 0x00496700,
        "last": 0x004967FF,
        "registers": [
            {
                "name": "SpotBatteryLoad",
                "offset": 62,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "SpotBatteryLoad",
                    Identifier.battery_charge_total,
                    unit="kWh",
                    factor=1000,
                ),
            }
        ],
    },
    "SpotBatteryUnload": {
        "command": 0x54000200,
        "response": 0x00496800,
        "first": 0x00496800,
        "last": 0x004968FF,
        "registers": [
            {
                "name": "SpotBatteryUnLoad",
                "offset": 62,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "SpotBatteryUnLoad",
                    Identifier.battery_discharge_total,
                    unit="kWh",
                    factor=1000,
                ),
            }
        ],
    },
    "EM_1": {
        "command": 0x54000200,
        "response": 0x00462400,
        "first": 0x00462400,
        "last": 0x004628FF,
        "registers": [
            {
                "name": "meter_yield",
                "offset": 66,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "meter_yield",
                    Identifier.metering_total_yield,
                    unit="kWh",
                    factor=1000,
                ),
            },
            {
                "name": "meter_absorbed",
                "offset": 78,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "meter_absorbed",
                    Identifier.metering_total_absorbed,
                    unit="kWh",
                    factor=1000,
                ),
            }

        ],
    },
    "EM_2": {
        "command": 0x54000200,
        "response": 0x40469100,
        "first": 0x40469100,
        "last": 0x404692FF,
    },
    "EM_3": {
        "command": 0x51000200,
        # "response": 0x40463600,
        # "first":    0x40463600,
        # "last": 0x404637FF,
        "response": 0x00463600,
        "first":    0x00463600,
        "last": 0x004637FF,
        "registers": [
            {
                "name": "meter_yield3",
                "offset": 62,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "meter_yield3",
                    Identifier.metering_power_supplied,
                    unit="W",
                    factor=1,
                ),
            },
            {
                "name": "meter_absorbed3",
                "offset": 90,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "meter_absorbed3",
                    Identifier.metering_power_absorbed,
                    unit="W",
                    factor=1,
                ),
            }

        ],
    },
    "EM_4": {
        "command": 0x51000200,
        "response": 0x0046E800,
        "first": 0x0046E800,
        "last": 0x0046EDFF,
    },
    "Insulation_1": {
        "command": 0x51020200,
        "response": 0x40254E00,
        "first": 0x40254E00,
        "last": 0x40254FFF,
    },
    "Insulation_2": {
        "command": 0x51020200,
        "response": 0x00254F00,
        "first": 0x00254F00,
        "last": 0x00254FFF,
    },
    "Firmware": {
        "command": 0x58000200,
        "response": 0x00823400,
        "first": 0x00823400,
        "last": 0x008234FF,
        "registers": [
            {
                "name": "Firmware",
                "format": "version",
                "offset": 78,
                "invalid": 0x80000000,
            }
        ],
    },
    "energy_production": {
        "command": 0x54000200,
        "response": 0x54000201,
        "first": 0x00260100, # TODO Check length
        "last": 0x002622FF,
        "registers": [

            {
                "name": "total",
                "offset": 62,
                "invalid": 0x8000000000000000,
                "sensor": Sensor(
                    "total", Identifier.total_yield, factor=1000, unit="kWh"
                ),
            },
            {
                "name": "today",
                "offset": 78,
                "invalid": 0x8000000000000000,
                "sensor": Sensor(
                    "today", Identifier.daily_yield, factor=1000, unit="kWh"
                ),
            },
        ],
        # $inv_SPOT_ETODAY,$inv_SPOT_ETOTAL
    },
    "spot_ac_power": {
        "command": 0x51000200,
        "response": 0x00464000,
        "first": 0x00464000,
        "last": 0x004642FF,
        "registers": [
            {
                "name": "spot_ac_power_l1",
                "offset": 62,
                "format": "int",
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "spot_ac_power1", Identifier.power_l1, factor=1, unit="W"
                ),
            },
            {
                "name": "spot_ac_power_l2",
                "offset": 90,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "spot_ac_power2", Identifier.power_l2, factor=1, unit="W"
                ),
            },
            {
                "name": "spot_ac_power_l3",
                "offset": 118,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "spot_ac_power3", Identifier.power_l3, factor=1, unit="W"
                ),
            },
        ],
    },
    "spot_ac_voltage": {
        "command": 0x51000200,
        "response": 0x00464800,
        "first": 0x00464800,
        "last": 0x004655FF,
        "registers": [
            {
                "name": "spot_ac_voltage1",
                "offset": 62,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "spot_ac_voltage1", Identifier.voltage_l1, factor=100, unit="V"
                ),
            },
            {
                "name": "spot_ac_voltage2",
                "offset": 90,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "spot_ac_voltage2", Identifier.voltage_l2, factor=100, unit="V"
                ),
            },
            {
                "name": "spot_ac_voltage3",
                "offset": 118,
                "invalid": 0x80000000,
                "sensor": Sensor(
                    "spot_ac_voltage3", Identifier.voltage_l3, factor=100, unit="V"
                ),
            },
            # {
            #     "name": "spot_ac_current1",
            #     "offset": 146,
            #     "invalid": 0x80000000,
            #     "format": "int",
            #     "sensor": Sensor(
            #         "spot_ac_current1", Identifier.current_l1, factor=100, unit="A"
            #     ),
            # },
            # {
            #     "name": "spot_ac_current2",
            #     "offset": 174,
            #     "invalid": 0x80000000,
            #     "format": "int",
            #     "sensor": Sensor(
            #         "spot_ac_current2", Identifier.current_l2, factor=100, unit="A"
            #     ),
            # },
            # {
            #     "name": "spot_ac_current3",
            #     "offset": 202,
            #     "invalid": 0x80000000,
            #     "format": "int",
            #     "sensor": Sensor(
            #         "spot_ac_current3", Identifier.current_l3, factor=100, unit="A"
            #     ),
            # },
        ],
    },
}



#based on https://github.com/Wired-Square/sma-query/blob/main/src/sma_query_sw/protocol.py
import logging
import time
import ctypes
import binascii
import copy
import collections
import struct

from typing import Any, Dict, Optional,List
from ctypes import LittleEndianStructure
from asyncio import DatagramProtocol

from .helpers import version_int_to_string
from .definitions_speedwire import commands

from .sensor import Sensor
from .const import Identifier
from .const import SMATagList

from .exceptions import (
    SmaConnectionException,
    SmaReadException,
    SmaAuthenticationException
)
from .sensor import Sensors
from .device import Device
import asyncio

_LOGGER = logging.getLogger(__name__)

APP_ID = 125
ANY_SERIAL = 0xFFFFFFFF
ANY_SUSYID = 0xFFFF

# Login Timeout in seconds
LOGIN_TIMEOUT = 900

# Create a reverse index based command lookup
ril_index = {f"ril-{value['first']:X}": value for (key, value) in commands.items() if "first" in value}


def get_encoded_pw(password, installer=False):
    # user=0x88, installer=0xBB
    byte_password = bytearray(password.encode("ascii"))

    if installer:
        login_code = 0xBB
    else:
        login_code = 0x88

    encodedpw = bytearray(12)

    for index in range(0, 12):
        if index < len(byte_password):
            encodedpw[index] = (login_code + byte_password[index]) % 256
        else:
            encodedpw[index] = login_code

    return encodedpw


class SpeedwireFrame:
    _frame_sequence = 1
    _id = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"SMA\x00"))
    _tag0 = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"\x00\x04\x02\xA0"))
    _group1 = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"\x00\x00\x00\x01"))
    _eth_sig = (ctypes.c_ubyte * 4).from_buffer(bytearray(b"\x00\x10\x60\x65"))
    _ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x01"))
    _ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x01"))

    _data_length = 0  # Placeholder value
    _longwords = 0  # Placeholder value
    _ctrl = 0  # Placeholder value

    class FrameHeader(LittleEndianStructure):
        _pack_ = 1
        _fields_ = [
            ("id", ctypes.c_ubyte * 4),
            ("tag0", ctypes.c_ubyte * 4),
            ("group1", ctypes.c_ubyte * 4),
            ("data_length", ctypes.c_uint16),
            ("eth_sig", ctypes.c_ubyte * 4),
            ("longwords", ctypes.c_ubyte),
            ("ctrl", ctypes.c_ubyte),
        ]

    class DataHeader(LittleEndianStructure):
        _pack_ = 1
        _fields_ = [
            ("dst_sysyid", ctypes.c_uint16),
            ("dst_serial", ctypes.c_uint32),
            ("ctrl2_1", ctypes.c_ubyte * 2),
            ("app_id", ctypes.c_uint16),
            ("app_serial", ctypes.c_uint32),
            ("ctrl2_2", ctypes.c_ubyte * 2),
            ("preamble", ctypes.c_uint32),
            ("sequence", ctypes.c_uint16),
        ]

    class LogoutFrame(LittleEndianStructure):
        _pack_ = 1
        _fields_ = [
            ("command", ctypes.c_uint32),
            ("data_start", ctypes.c_uint32),
            ("data_end", ctypes.c_uint32),
        ]

    class LoginFrame(LittleEndianStructure):
        _pack_ = 1
        _fields_ = [
            ("command", ctypes.c_uint32),
            ("login_type", ctypes.c_uint32),
            ("timeout", ctypes.c_uint32),
            ("time", ctypes.c_uint32),
            ("data_start", ctypes.c_uint32),
            ("user_password", ctypes.c_ubyte * 12),
            ("data_end", ctypes.c_uint32),
        ]

    class QueryFrame(LittleEndianStructure):
        _pack_ = 1
        _fields_ = [
            ("command", ctypes.c_uint32),
            ("first", ctypes.c_uint32),
            ("last", ctypes.c_uint32),
            ("data_end", ctypes.c_uint32)
        ]

    def getLogoutFrame(self, inverter):
        frame_header = self.getFrameHeader()
        frame_data_header = self.getDataHeader(inverter)
        frame_data = self.LogoutFrame()

        frame_header.ctrl = 0xA0
        frame_data_header.dst_sysyid = 0xFFFF
        frame_data_header.ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x03"))
        frame_data_header.ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x03"))

        frame_data.command = commands["logoff"]["command"]
        frame_data.data_start = 0xFFFFFFFF
        frame_data.data_end = 0x00000000

        data_length = ctypes.sizeof(frame_data_header) + ctypes.sizeof(frame_data)

        frame_header.data_length = int.from_bytes(data_length.to_bytes(2, "big"), "little")

        frame_header.longwords = (data_length // 4)

        return bytes(frame_header) + bytes(frame_data_header) + bytes(frame_data)

    def getLoginFrame(self, inverter, installer):
        frame_header = self.getFrameHeader()
        frame_data_header = self.getDataHeader(inverter)
        frame_data = self.LoginFrame()

        frame_header.ctrl = 0xA0
        frame_data_header.dst_sysyid = 0xFFFF
        frame_data_header.ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x01"))
        frame_data_header.ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x01"))

        frame_data.command = commands["login"]["command"]
        frame_data.login_type = (0x07, 0x0A)[installer]
        frame_data.timeout = LOGIN_TIMEOUT
        frame_data.time = int(time.time())
        frame_data.data_start = 0x00000000  # Data Start
        frame_data.user_password = (ctypes.c_ubyte * 12).from_buffer(get_encoded_pw(inverter["user_password"], installer))
        frame_data.date_end = 0x00000000 # Packet End

        data_length = ctypes.sizeof(frame_data_header) + ctypes.sizeof(frame_data)

        frame_header.data_length = int.from_bytes(data_length.to_bytes(2, "big"), "little")

        frame_header.longwords = (data_length // 4)

        return bytes(frame_header) + bytes(frame_data_header) + bytes(frame_data)

    def getQueryFrame(self, inverter, command_name):
        frame_header = self.getFrameHeader()
        frame_data_header = self.getDataHeader(inverter)
        frame_data = self.QueryFrame()

        command = commands[command_name]

        frame_header.ctrl = 0xA0
        frame_data_header.dst_sysyid = 0xFFFF
        frame_data_header.ctrl2_1 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x00"))
        frame_data_header.ctrl2_2 = (ctypes.c_ubyte * 2).from_buffer(bytearray(b"\x00\x00"))

        frame_data.command = command["command"]
        frame_data.first = command["first"]
        frame_data.last = command["last"]
        frame_data.date_end = 0x00000000

        data_length = ctypes.sizeof(frame_data_header) + ctypes.sizeof(frame_data)

        frame_header.data_length = int.from_bytes(data_length.to_bytes(2, "big"), "little")

        frame_header.longwords = (data_length // 4)

        return bytes(frame_header) + bytes(frame_data_header) + bytes(frame_data)

    def getFrameHeader(self):
        newFrameHeader = self.FrameHeader()
        newFrameHeader.id = self._id
        newFrameHeader.tag0 = self._tag0
        newFrameHeader.group1 = self._group1
        newFrameHeader.data_length = self._data_length
        newFrameHeader.eth_sig = self._eth_sig
        newFrameHeader.longwords = self._longwords
        newFrameHeader.ctrl = self._ctrl

        return newFrameHeader

    def getDataHeader(self, inverter):
        newDataHeader = self.DataHeader()

        newDataHeader.dst_susyid = ANY_SUSYID
        newDataHeader.dst_serial = ANY_SERIAL
        newDataHeader.ctrl2_1 = self._ctrl2_1
        newDataHeader.app_id = APP_ID
        newDataHeader.app_serial = (inverter["serial"])
        newDataHeader.ctrl2_2 = self._ctrl2_2
        newDataHeader.preamble = 0
        newDataHeader.sequence = (self._frame_sequence | 0x8000)

        self._frame_sequence += 1

        return newDataHeader


class SMAClientProtocol(DatagramProtocol):

    useDummyData = True    
    debug = {
        "msg": collections.deque(maxlen=30),
        "data": {},
        "unfinished": set()
    }

    def __init__(self, inverter, on_connection_lost):
        self.speedwire = SpeedwireFrame()
        self.transport = None
        self.inverter = inverter
        self.on_connection_lost = on_connection_lost
        self.cmds = []
        self.cmdidx = 0
        self.future = None

        self.allCmds = []
        self.allCmds.extend(commands.keys())
        self.allCmds.remove("login")
        self.allCmds.remove("logoff")

    def connection_made(self, transport):
        self.transport = transport

    def start_query(self, cmds: List, future, group: str):
        self.cmds = cmds
        self.future = future
        self.cmdidx = 0
        self.inverter["data"] = {}
        self.inverter["sensors"] = {}
        _LOGGER.debug(f"""Sending login for {self.inverter["serial"]:#08X}""")
        if self.useDummyData:
            self.inverter["data"] = {}
            self.inverter["sensors"] = {}
            self.cmds = []
            self._send_next_command()
        else:
            groupidx = ["user", "installer"].index(group)
            self._send_command(self.speedwire.getLoginFrame(self.inverter, groupidx))
    
    def connection_lost(self, exc):
        _LOGGER.debug(f"Connection lost: {exc}")
        self.on_connection_lost.set_result(True)

    def _send_command(self, cmd):
        _LOGGER.debug(f"Sending command [{len(cmd)}] -- {binascii.hexlify(cmd).upper()}")
        self.transport.sendto(cmd)

    def _get_code(self, data):
        code = int.from_bytes(data[42:46], "little")
        return code

    def _get_ril(self, data):
        code = int.from_bytes(data[54:58], "little")
        return code & 0x00FFFF00

    def _get_value(self, data, offset, format, invalid):
        if format == "int":
            format = "<l"
        elif format == "uint":
            format = "<L"
        size = struct.calcsize(format)
        if (size + offset >= len(data)):
            return None
        v = struct.unpack(format, data[offset:offset+size])[0]
        if (v in [0x80000000, 0xFFFFFFFF, -0x80000000 ] or (invalid and invalid == v)):
            v = None
        return v

    def _get_long_value(self, data, offset):
         value = int.from_bytes(data[offset:offset + 4], "little")
         return value

    def _send_next_command(self):
        if not self.future:
            return
        if self.cmdidx >= len(self.cmds):
            if self.useDummyData:
                f = self.future
                self.future = None
                self.addDummyData()
                self.future = f
            self.debug["data"] = self.inverter["data"]
            self.future.set_result(True)
            self.cmds = []
            self.cmdidx = 0
            self.future = None
        else:
            self.debug["msg"].append(["SEND", self.cmds[self.cmdidx]])
            _LOGGER.debug("Sending "+ self.cmds[self.cmdidx])
            self._send_command(self.speedwire.getQueryFrame(self.inverter,
                                               self.cmds[self.cmdidx]))
            self.cmdidx += 1

    def _getvalue(self, register, data):
            format = register.get("format", "uint")
            value = None
            if format == "list":
                i = register.get("offset")
                while (int.from_bytes(data[i:i+4], byteorder="little") != 0x00FFFFFE) and i+4 <= len(data):
                        temp = int.from_bytes(data[i:i+4], byteorder="little")
                        if (temp & 0xFF000000) == 0x01000000:
                                value = temp & 0x00FFFFFF
                        i += 4
            elif format == "version":
#                        value = self._get_value(data, register["offset"], format, register.get("invalid"))
                value = self._get_long_value(data, register["offset"])
                value = version_int_to_string(value)
            else:
                if register["offset"] >= len(data):
                    _LOGGER.debug(f'Value for {register["name"]} not found in this message type')
                    return None
                value = self._get_value(data, register["offset"], format, register.get("invalid"))
                if value and register.get("mask"):
                    value = value & register.get("mask")
            return value

    def datagram_received(self, data, addr):
        code = self._get_code(data)
        ril = self._get_ril(data)
        ril_key = f"ril-{ril:X}"
        shortcode = (f"{ril:X}")[0:4]

        _LOGGER.debug(f"{addr} -- [{len(data)}] [{code:X}] [{ril:X}] -- {binascii.hexlify(data).upper()}")
        self.debug["msg"].append(["RECV", len(data), (f"{ril:X}")[0:4], binascii.hexlify(data).upper().decode("utf-8")])

        if code == commands["login"]["response"]:
            _LOGGER.debug("Login repsonse received!")
            self.inverter["data"] = {}
            self.inverter["sensors"] = {}
            for r in commands["login"]["registers"]:
                value = self._getvalue(r, data)
                self.inverter["data"][r['name']] = value
            if (self.inverter["data"]["error"] == 256):
                _LOGGER.error("Login failed!")
                self.future.set_exception(SmaAuthenticationException("Login failed! Credentials wrong (user/install or password)"))

        if len(data) <= 58:
            _LOGGER.debug(f"Short datagram received: [{len(data)}] -- {data}")
            self._send_next_command()
            return
        
        if ril_key not in ril_index:
            shortcode = (f"{ril:X}")[0:4]
            _LOGGER.debug(f"No Definition for {ril_key} {code:X} {shortcode}")
            self.debug["unfinished"].add(f"No Definition for {ril_key} {code:X} {shortcode}")
            self._send_next_command()
            return
        
        if ril_key in ril_index:
            command = ril_index[ril_key]
            short = f"{command['first']:#08X}"[2:6]
            if "registers" in command:
                if len(command["registers"]) == 0:
                    self.debug["unfinished"].add(f"Command Response: {ril_key} {short} {code:X}")
                    _LOGGER.error(f"Message but no registers definied. {command}")
                for register in command["registers"]:
                    value = self._getvalue(register, data)
                    if value is not None:
                        if "sensor" in register:
                            sen = copy.copy(register["sensor"])
                            v = value
                            if (sen.factor and sen.factor != 1):
                                v /= sen.factor
                            sen.value = v
                            self.inverter["sensors"][sen.key] = sen
                        else:
                             _LOGGER.debug(f'Value but no sensor: {register["name"]} {value}')
                             pass
                    self.inverter["data"][register["name"]] = value
            else:
                if self.useDummyData:
                    _LOGGER.debug(f" {shortcode} [{len(data)}] [{code:X}] [{ril:X}]")
                    print("======================================================================================================")
                    print("======================================================================================================")
                    raise RuntimeError()
            self._send_next_command()

    def addDummyData(self):
        for pp in []:
            self.datagram_received(bytes.fromhex(pp), ('10.10.10.10', 9522))


class SMAspeedwireINV(Device):

    _transport = None
    _protocol = None
    _deviceinfo = {}

    def __init__(
        self,
        host: str,
        group: str,
        password: Optional[str]
    ):
        self._host = host
        self._group = group
        if group not in ["user", "installer"]:
            raise KeyError(f"Invalid user type: {group} (user or installer)")

        self._inverter = {
            "serial": 0,
            "user_password": password,
        }
    #     self.check()

    # def check(self) -> None:
    #     keysname = {}
    #     sensorname = {}
    #     for x in commands.items():
    #         if "registers" in x[1]:
    #             for r in x[1]["registers"]:
    #                 if "name" in r:
    #                     name = r["name"]
    #                     if (name in keysname):
    #                         print("Doppelter Keyname " + name)
    #                         raise RuntimeError("Doppelter Keyname " + name)
    #                     keysname[name] = 1
    #                 if "sensor" in r:
    #                     sensor = r["sensor"]
    #                     if (sensor.key in sensorname):
    #                         print("Doppelter Sensorname " + sensor.key)
    #                         raise RuntimeError("Doppelter Sensorname " + sensor.key)

    #                     sensorname[name] = 1


    async def get_sensors(self) -> Sensors:
        fut = asyncio.get_running_loop().create_future()
        c = self._protocol.allCmds
        if self._protocol.useDummyData:
            c = []
        self._protocol.start_query(c, fut, self._group)
        await asyncio.wait_for(fut, timeout=5)
        device_sensors = Sensors()
        for s in self._inverter["sensors"].values():
            device_sensors.add(s)
        return device_sensors
    
    async def new_session(self) -> bool:
        loop = asyncio.get_running_loop()
        on_connection_lost = loop.create_future()

        self._transport, self._protocol = await loop.create_datagram_endpoint(
             lambda: SMAClientProtocol(self._inverter, on_connection_lost), 
             remote_addr=(self._host, 9522))

    async def device_info(self) -> dict:
        fut = asyncio.get_running_loop().create_future()
        self._protocol.start_query(["TypeLabel", "Firmware"], fut, self._group)
        try:
            await asyncio.wait_for(fut, timeout=5)
        except TimeoutError:
            if "error" in self._protocol.inverter["data"] and self._protocol.inverter["data"]["error"] == 0:
                raise SmaReadException("Reply for request not received")
            else:
                raise SmaConnectionException("No connection to device")

        data = self._inverter["data"]

        invcnr = self._inverter["data"].get("inverter_class",0)
        invc = SMATagList.get(invcnr, "Unknown device")

        invtnr = self._inverter["data"].get("inverter_type",0)
        invt = SMATagList.get(invtnr, "Unknown type")
        self._deviceinfo = {
            "serial": data.get("serial", ""),
            "name": str(invt) + " (" + str(invtnr) + ")",
            "type": str(invc) + " (" + str(invcnr) + ")",
            "manufacturer": "SMA",
            "sw_version": data.get("Firmware",""),
        }
        return self._deviceinfo


    async def read(self, sensors: Sensors) -> bool:
        fut = asyncio.get_running_loop().create_future()
        c = self._protocol.allCmds
        if self._protocol.useDummyData:
            c = []
        self._protocol.start_query(c, fut, self._group)
        await asyncio.wait_for(fut, timeout=5)
        sensorReadings = self._inverter["sensors"]
        for sen in sensors:
            if sen.enabled:
                if sen.key in sensorReadings:
                    value = sensorReadings[sen.key].value
                    if (sen.mapper):
                        sen.mapped_value = sen.mapper.get(value, str(value))
                    sen.value = value


    async def close_session(self) -> None:
        pass


    async def get_debug(self) -> Dict:
        return {
            "msg": list(self._protocol.debug["msg"]),
            "data": self._protocol.debug["data"],
            "device_info": self._deviceinfo,
            "unfinished": list(self._protocol.debug["unfinished"])
       }

